
# THIS FILE IS GENERATED FROM wofry SETUP.PY
short_version = '1.0.23'
version = '1.0.23'
full_version = '1.0.23'
git_revision = '23c649e9fa948214b2feabe84ea0306fc42638de'
release = True

if not release:
    version = full_version
    short_version += ".dev"
